/*
 * com/ngon/tap/ClusStats.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.gfx.RGB;
import com.ngon.math.Gradient2D;
import com.ngon.math.Math2D;
import com.ngon.mpeg.MPEGFrame;
import com.ngon.mpeg.MPEGFrameStats;
import java.awt.Dimension;
import java.awt.Point;
import java.util.Enumeration;
import java.util.Vector;

/**
 * class ClusStats
 */
final class ClusStats extends Vector {
    
    final static int NEIGHBOR_RADIUS        = 2;
    final static int NEIGHBOR_THRESHOLD     = 100;
    final static int STATIONARY_THRESHOLD   = 1;
    final static int ANGLE_THRESHOLD        = 40;
    
    final static int SPURIOUS_EDGE_DIST     = 16;
    final static int SPURIOUS_BLOCK_SZ      = 6;
    final static int SPURIOUS_MV_LEN        = 16;
    
    final static int GRAD_BITSHIFT         = 1;
    
    private MPEGFrame frame;
    private Point[] mv;
    private Gradient2D[] mvGrad;
    
    ClusStats(MPEGFrame frame) {
        this.frame = frame;
    }
    
    /**
     * Run through blocks of MPEGFrame, 
     * adding neighbors sharing similar motion.
     */
    static public Vector findClusters(MPEGFrame frame) {
        
        Vector[] neighbors = findNeighbors(frame);
        
        // Aggregate neighbors into clusters of similar motion.  Do this by
        // gathering neighbors of a given block, then adding the neighbors 
        // of these neighbors, etc.
        Vector clusters = new Vector();
        for (int idx = 0; idx < neighbors.length; idx++) {
            
            if (neighbors[idx] == null) {
                continue;
            }
            
            // Check that no cluster already contains this block.
            if (containingCluster(idx, clusters) != null) {
                continue;
            }
            
            // Ok, passed checks.  Create a new ClusStats object.
            ClusStats c = new ClusStats(frame);
            c.seed(new Integer(idx), neighbors);
            
            // Remove spurious clusters.
//            if (c.isSpurious()) {
//                continue;
//            }
            
            clusters.addElement(c);
        }
        
        clusters.trimToSize();
        return clusters;
    }
    
    /**
     * Run through blocks of MPEGFrame, 
     * finding neighborhoods of similar motion.
     * Return an array of neighborhoods (each a Vector of blocks)
     */
    static Vector[] findNeighbors(MPEGFrame frame) {
        
        MPEGFrameStats stats = frame.getStats();
        if (stats == null) {
            System.err.println("No stats in MPEGFrame!");
            return null;
        }
        if (! stats.isRecordingStat(MPEGFrameStats.DIMENSIONS) ||
            ! stats.isRecordingStat(MPEGFrameStats.MOTION_VECTORS)) {
            System.err.println(
                "Must be recording stats DIMENSIONS and MOTION_VECTORS");
            return null;
        }
        
        Dimension d = stats.getMacroblockDimensions();
        Point[] mvField = stats.getMotionVectors();
        
        int n = d.width * d.height;
        Vector[] neighbors = new Vector[n];
        for (int j = 0; j < d.height; j++) {
            for (int i = 0; i < d.width; i++) {
                int idx0 = i + j * d.width;
                neighbors[idx0] = null;
                Point mv0 = mvField[idx0];
                if (mv0.x * mv0.x + mv0.y * mv0.y < STATIONARY_THRESHOLD) {
                    // Omit stationary blocks.
                    continue;
                }
                int p, q;
                for (q = j-NEIGHBOR_RADIUS; q <= j+NEIGHBOR_RADIUS; q++) {
                    if (q < 0 || q >= d.height) {
                        continue;
                    }
                    for (p = i-NEIGHBOR_RADIUS; p <= i+NEIGHBOR_RADIUS; p++) {
                        if (p < 0 || p >= d.width) {
                            continue;
                        }
                        if (p == i && q == j) {
                            continue;
                        }
                        int idx = p + q * d.width;
                        Point mv = mvField[idx];
                        if (mv.x * mv.x + mv.y * mv.y < STATIONARY_THRESHOLD) {
                            continue;
                        }
                        int dx = mv.x - mv0.x;
                        int dy = mv.y - mv0.y;
                        if (dx * dx + dy * dy > NEIGHBOR_THRESHOLD) {
                            continue;
                        }
                        if (Math.abs(Math2D.angleBetween(mv, mv0)) >
                            ANGLE_THRESHOLD) {
                            continue;
                        }
                        // Passed all checks.
                        if (neighbors[idx0] == null) {
                            neighbors[idx0] = new Vector();
                        }
                        neighbors[idx0].addElement(new Integer(idx));
                    }
                }
            }
        }
        
        return neighbors;
    }
    
    static public ClusStats containingCluster(int idx, Vector clusters) {
        for (Enumeration e = clusters.elements(); e.hasMoreElements(); ) {
            ClusStats c = (ClusStats) e.nextElement();
            if (c.contains(idx)) {
                return c;
            }
        }
        return null;
    }
    
    private void seed(Integer o, Vector[] neighbors) {
        if (this.contains(o)) {
            return;
        }
        addElement(o);
        int i = o.intValue();
        if (neighbors[i] == null) {
            return;
        }
        for (Enumeration e = neighbors[i].elements(); e.hasMoreElements(); ) {
            Integer neighbor = (Integer) e.nextElement();
            seed(neighbor, neighbors);
        }
    }
    
    public Point getAverageMV() {
        int sumX = 0;
        int sumY = 0;
        int n = size();
        Point[] mvField = getMotionVectors();
        for (int i = 0; i < n; i++) {
            int idx = ((Integer) elementAt(i)).intValue();
            sumX += mvField[idx].x;
            sumY += mvField[idx].y;
        }
        return new Point(sumX/n, sumY/n);
    }
    
    public Point getCenterOfGravity() {
        int sumX = 0;
        int sumY = 0;
        int n = size();
        Dimension d = getFieldDimensions();
        for (int i = 0; i < n; i++) {
            int idx = ((Integer) elementAt(i)).intValue();
            sumX += (idx % d.width) * 16 + 7;
            sumY += (idx / d.width) * 16 + 7;
        }
        return new Point(sumX/n, sumY/n);
    }
    
    public int getAverageDiv() {
        int sum = 0;
        int n = size();
        for (int i = 0; i < n; i++) {
            Integer o = (Integer) elementAt(i);
            Gradient2D grad2D = getGradient(o);
            sum += Gradient2D.div(grad2D);
        }
        return sum/n;
    }
    
    public int getAverageCurl() {
        int sum = 0;
        int n = size();
        for (int i = 0; i < n; i++) {
            Integer o = (Integer) elementAt(i);
            Gradient2D grad2D = getGradient(o);
            sum += Gradient2D.curl(grad2D);
        }
        return sum/n;
    }
    
    public Gradient2D getGradient(Integer o) {
        int idx0 = o.intValue();
        Point[] mv = getMotionVectors();
        Dimension d = getFieldDimensions();
        Gradient2D grad2D = new Gradient2D();
        int x = idx0 % d.width;
        int y = idx0 / d.height;
        int count = 0;
        if (x > 0) {
            int idx = idx0 - 1;
            if (contains(new Integer(idx))) {
                grad2D.dVx.x += (mv[idx0].x - mv[idx].x) << GRAD_BITSHIFT;
                grad2D.dVy.x += (mv[idx0].y - mv[idx].y) << GRAD_BITSHIFT;
                count++;
            }
        }
        if (x < d.width - 1) {
            int idx = idx0 + 1;
            if (contains(new Integer(idx))) {
                grad2D.dVx.x += (mv[idx].x - mv[idx0].x) << GRAD_BITSHIFT;
                grad2D.dVy.x += (mv[idx].y - mv[idx0].y) << GRAD_BITSHIFT;
                count++;
            }
        }
        if (count > 1) {
            grad2D.dVx.x /= count;
            grad2D.dVy.x /= count;
        }
        count = 0;
        if (y > 0) {
            int idx = idx0 - d.width;
            if (contains(new Integer(idx))) {
                grad2D.dVx.y += (mv[idx0].x - mv[idx].x) << GRAD_BITSHIFT;
                grad2D.dVy.y += (mv[idx0].y - mv[idx].y) << GRAD_BITSHIFT;
                count++;
            }
        }
        if (x < d.width - 1) {
            int idx = idx0 + 1;
            if (contains(new Integer(idx))) {
                grad2D.dVx.y += (mv[idx].x - mv[idx0].x) << GRAD_BITSHIFT;
                grad2D.dVy.y += (mv[idx].y - mv[idx0].y) << GRAD_BITSHIFT;
                count++;
            }
        }
        if (count > 1) {
            grad2D.dVx.y /= count;
            grad2D.dVy.y /= count;
        }
        return grad2D;
    }
    
    public int getAverageRGB() {
        long sumR = 0;
        long sumG = 0;
        long sumB = 0;
        int n = size();
        Dimension d = getFieldDimensions();
        int[] rgbField = frame.getRGB();
        for (int i = 0; i < n; i++) {
            int idx = ((Integer) elementAt(i)).intValue();
            int x0 = (idx % d.width) * 16;
            int y0 = (idx / d.width) * 16;
            for (int y = y0; y < y0 + 16; y++) {
                for (int x = x0; x < x0 + 16; x++) {
                    int rgb = rgbField[x + y * d.width * 16];
                    sumR += (rgb & RGB.RED) >> 16;
                    sumG += (rgb & RGB.GRN) >>  8;
                    sumB += (rgb & RGB.BLU);
                }
            }
        }
        return (((int) ((sumR / n) << 8)) & RGB.RED) | 
               (((int)  (sumG / n)      ) & RGB.GRN) |
               (((int) ((sumB / n) >> 8)) & RGB.BLU) ;
    }
    
    public Dimension getClusterDimensions() {
        int w = 0;
        int h = 0;
        int n = size();
        if (n > 0) {
            Dimension d = getFieldDimensions();
            int xMin = d.width;
            int xMax = 0;
            int yMin = d.height;
            int yMax = 0;
            for (int i = 0; i < n; i++) {
                int idx = ((Integer) elementAt(i)).intValue();
                int x = idx % d.width;
                int y = idx / d.width;
                xMin = Math.min(x, xMin);
                yMin = Math.min(y, yMin);
                xMax = Math.max(x, xMax);
                yMax = Math.max(y, yMax);
            }
            w = Math.max(xMax - xMin + 1, 0);
            h = Math.max(yMax - yMin + 1, 0);
        }
        return new Dimension(w, h);
    }
    
    public boolean contains(int idx) {
        return contains(new Integer(idx));
    }
    
    /**
     * A spurious cluster happens near the edge (<= SPURIOUS_EDGE_DIST),
     * has few blocks (<= SPURIOUS_BLOCK_SZ), and has large motion vectors
     * (>= SPURIOUS_MV_LEN).
     */
    public boolean isSpuriousOld() {
        if (size() > SPURIOUS_BLOCK_SZ) {
            return false;
        }
        Point mv = getAverageMV();
        if (mv.x * mv.x + mv.y * mv.y < SPURIOUS_MV_LEN * SPURIOUS_MV_LEN) {
            return false;
        }
        Point cg = getCenterOfGravity();
        Dimension d = getFieldDimensions();
        if (cg.x > SPURIOUS_EDGE_DIST &&
            cg.x < d.width * 16 - SPURIOUS_EDGE_DIST && 
            cg.y > SPURIOUS_EDGE_DIST &&
            cg.y < d.height * 16 - SPURIOUS_EDGE_DIST) {
            return false;
        }
        return true;
    }
    
    /**
     * A spurious cluster happens near the edge (<= SPURIOUS_EDGE_DIST)
     * and points the wrong way.
     */
    public boolean isSpurious() {
        Dimension d = getFieldDimensions();
        if (size() > d.width) {
            return false;
        }
        Point mv = getAverageMV();
        Point cg = getCenterOfGravity();
        if ((cg.x <= SPURIOUS_EDGE_DIST && -mv.x <= 0) ||
            (cg.y <= SPURIOUS_EDGE_DIST && -mv.y <= 0) || 
            (cg.x >= d.width*16 - SPURIOUS_EDGE_DIST && -mv.x >= 0) ||
            (cg.y >= d.height*16 - SPURIOUS_EDGE_DIST && -mv.y >= 0)) {
            return true;
        }
        return false;
    }
    
    private Dimension getFieldDimensions() {
        MPEGFrameStats stats = frame.getStats();
        if (stats == null) {
            System.err.println("No stats in MPEGFrame!");
            return null;
        }
        if (! stats.isRecordingStat(MPEGFrameStats.DIMENSIONS)) {
            System.err.println("Must be recording stats DIMENSIONS");
            return null;
        }
        return stats.getMacroblockDimensions();
    }
    
    private Point[] getMotionVectors() {
        if (mv != null) {
            return mv;
        }
        MPEGFrameStats stats = frame.getStats();
        if (stats == null) {
            System.err.println("No stats in MPEGFrame!");
            return null;
        }
        if (! stats.isRecordingStat(MPEGFrameStats.MOTION_VECTORS)) {
            System.err.println("Must be recording stats MOTION_VECTORS");
            return null;
        }
        mv = stats.getMotionVectors();
        return mv;
    }
    
    private Gradient2D[] getMotionVectorGradients() {
        if (mvGrad != null) {
            return mvGrad;
        }
        Point[] mv = getMotionVectors();
        Dimension d = getFieldDimensions();
        mvGrad = Gradient2D.gradient2D(mv, d);
        return mvGrad;
    }
}

